/*
* dll.c
  Copyright Jo'zsef Ne'meth 2003, joe.nemeth@cpluscsystems.axelero.net
*/
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef WIN32_LEAN_AND_MEAN

#include <tclInt.h>

#ifndef DECLSPEC_EXPORT
   #define DECLSPEC_EXPORT __declspec(dllexport)
#endif /* DECLSPEC_EXPORT */

BOOL APIENTRY
DllMain(HANDLE hModule, DWORD dwReason, LPVOID lpReserved)
{
   return TRUE;
}

int dllLoadObjCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
int dllCreateCmdObjCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
int dllCallObjCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
int dllBufferObjCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
int dllRefObjCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
int dllDeRefObjCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
int dllMemRdObjCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
int dllMemWrObjCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

#define DLL "::dll"


#define VOIDPAR	0x00000000
#define CHAR		0x00000001
#define SHORT		0x00000002
#define INT		0x00000004
#define INT64		0x00000008
#define FLOAT		0x00000010
#define DOUBLE	0x00000020
#define LONGDOUBLE	0x00000020
#define UNSIGNED	0x00000040
#define NUMERIC	0x00000080
#define STRING	0x00000100
#define BUFRET	0x10000000
#define STRUCT	0x20000000
#define PTR		0x40000000
#define PTRPTR	0x80000000

/*
* Dll_Init --
*/

EXTERN_C int DECLSPEC_EXPORT
Dll_Init(Tcl_Interp* interp)
{
  /* Initialize the stub table interface */

   if (Tcl_InitStubs(interp, "8.4", 0) == NULL) {
      return TCL_ERROR;
   }

   Tcl_CreateObjCommand(interp, DLL "::load_module", dllLoadObjCmd, (ClientData) NULL, NULL);
   Tcl_CreateObjCommand(interp, DLL "::create_cmd", dllCreateCmdObjCmd, (ClientData) NULL, NULL);
   Tcl_CreateObjCommand(interp, DLL "::buffer", dllBufferObjCmd, (ClientData) NULL, NULL);
   Tcl_CreateObjCommand(interp, DLL "::ref", dllRefObjCmd, (ClientData) NULL, NULL);
   Tcl_CreateObjCommand(interp, DLL "::deref", dllDeRefObjCmd, (ClientData) NULL, NULL);
   Tcl_CreateObjCommand(interp, DLL "::memrd", dllMemRdObjCmd, (ClientData) NULL, NULL);
   Tcl_CreateObjCommand(interp, DLL "::memwr", dllMemWrObjCmd, (ClientData) NULL, NULL);

   Tcl_SetVar(interp, "::DLL", DLL, 0);

   Tcl_PkgProvide(interp, "dll", "1.0");
   interp->result = "dll 1.0, namespace: " DLL;
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 */

int
dllLoadObjCmd(clientData, interp, objc, objv)
   ClientData clientData;
   Tcl_Interp *interp;
   int objc;
   Tcl_Obj *CONST objv[];
{
   if (objc != 2) {
      Tcl_WrongNumArgs(interp, 1, objv, "dllFileName ? -> namespace?");
      return TCL_ERROR;
   }

   Tcl_SetObjResult(interp, Tcl_NewIntObj((int) LoadLibrary(Tcl_GetStringFromObj(objv[1], NULL))));
   return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 */


int
dllCreateCmdObjCmd(clientData, interp, objc, objv)
   ClientData clientData;
   Tcl_Interp *interp;
   int objc;
   Tcl_Obj *CONST objv[];
{
   int ord, handle;
   void *address;

   if (objc != 4) {
      Tcl_WrongNumArgs(interp, 1, objv, "cmd handle name");
      return TCL_ERROR;
   }
   Tcl_GetIntFromObj(interp, objv[2], &handle);

   if (Tcl_GetIntFromObj(interp, objv[3], &ord) == TCL_ERROR) {
      address = GetProcAddress((struct HINSTANCE__ *) handle, Tcl_GetStringFromObj(objv[3], NULL));
   }
   else {
      address = GetProcAddress((struct HINSTANCE__ *) handle, (char *) ord);
   }

   Tcl_CreateObjCommand(interp, Tcl_GetStringFromObj(objv[1], NULL), dllCallObjCmd, (ClientData) NULL, NULL);
   Tcl_SetObjResult(interp, Tcl_NewIntObj((int) address));
   return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 */

void dll_call (char * values, unsigned bytes,
               unsigned flags, unsigned * r_value,
               void (*fn)(), void *(copy)(void *, const void *, unsigned int))
{
   __asm {
 
     // Make room for all of the new args.
        mov   ecx, bytes                                                     
        sub   esp, ecx
 
        mov   esi, esp

     // Place all of the memcpy args in position
        push  ecx
        push  values
        push  esi
        call  copy
 
     // Return stack to previous state and call the function
        add   esp, 12

        call  fn
 
     // If necessary remove the space we pushed for the args
        cmp   esi, esp
        jne   cont
        mov   ecx, bytes
        add   esp, ecx
 
     // Load %ecx with the return type code
cont:
        mov  ecx, flags
 
     // If the return value pointer is NULL, assume no return value.
        cmp  r_value, 0
        jne   retint
 
     // Even if there is no space for the return value, we are
     // obliged to handle floating-point values.
        cmp   ecx, FLOAT
        jne   noretval
        fstp  st(0)
 
        jmp   epilogue
 
retint:
        cmp   ecx, INT
        jne   retfloat
     // Load %ecx with the pointer to storage for the return value
        mov   ecx, r_value
        mov   [ecx], eax
        jmp   epilogue
 
retfloat:
        cmp  ecx, FLOAT
        jne   retdouble
     // Load  ecx with the pointer to storage for the return value
        mov   ecx, r_value
        fstp  dword ptr [ecx]                                   ; ??
        jmp   epilogue
 
retdouble:
        cmp  ecx, DOUBLE
        jne   retlongdouble
     // Load  ecx with the pointer to storage for the return value
        mov   ecx, r_value
        fstp  qword ptr [ecx]                                   ; ??
        jmp   epilogue
 
retlongdouble:
        cmp   ecx, LONGDOUBLE
        jne   retint64
     // Load  ecx with the pointer to storage for the return value
        mov   ecx, r_value
        fstp  tbyte ptr [ecx]                                   ; ??
        jmp   epilogue
 
retint64:
        cmp   ecx, INT64
        jne   retstruct
     // Load  ecx with the pointer to storage for the return value
        mov   ecx, r_value
        mov   [ecx], eax
        mov   4[ecx], eax
 
retstruct:
     // Nothing to do!
 
noretval:
epilogue:
   }
}

/*
 *----------------------------------------------------------------------
 */

int
dllCallObjCmd(clientData, interp, objc, objv)
   ClientData clientData;
   Tcl_Interp *interp;
   int objc;
   Tcl_Obj *CONST objv[];
{
   int *params, *values, *valptr, *valstrt;
   int cnt, i, error_code;
   unsigned int type, ret_type;
   Tcl_Obj* tmp;
   char *vptr;

   params = (int *) Tcl_GetByteArrayFromObj(Tcl_GetVar2Ex(interp, DLL "::entry", Tcl_GetStringFromObj(objv[0], NULL), 0), NULL);
   cnt = params[2];
   values = params + cnt + 3;

   if (cnt != objc) {
      Tcl_WrongNumArgs(interp, 1, objv, ".............");
      return TCL_ERROR;
   }

   valstrt = values + 2;
   for (i = 1, valptr = valstrt; i < objc; i++) {
      type = params[i + 3];
      if (type < PTR) {
         if (type & NUMERIC) {
            if (type & (FLOAT || DOUBLE)) {
               if (Tcl_GetDoubleFromObj(interp, objv[i], ((double *) valptr)++) == TCL_ERROR) {
                  Tcl_SetResult(interp, "floating parameter error", NULL);
                  return TCL_ERROR;
               }
            }
            else {
               if (Tcl_GetIntFromObj(interp, objv[i], valptr++) == TCL_ERROR) {
                  Tcl_SetResult(interp, "integer parameter error", NULL);
                  return TCL_ERROR;
               }
            }
         }
         else if (type & STRING) {
            *valptr++ = (int) Tcl_GetString(objv[i]);
         }
         else {
            Tcl_SetResult(interp, "void may not be an input parameter", NULL);
            return TCL_ERROR;
         }
      }
      else {
         if ((tmp = Tcl_ObjGetVar2(interp, objv[i], NULL, 0)) == NULL) {
            if (Tcl_GetIntFromObj(interp, objv[i], valptr) == TCL_OK) {
               if (*valptr++ == 0) {
                  continue;
               }
               Tcl_SetResult(interp, "variable name or NULL is required for pointer", NULL);
               return TCL_ERROR;
            }
            tmp = Tcl_ObjSetVar2(interp, objv[i], NULL, Tcl_NewIntObj(0), 0);
         }
         if (type & NUMERIC) {
            if (type & (FLOAT || DOUBLE)) {
               if (Tcl_GetDoubleFromObj(interp, tmp, (double *) valptr) == TCL_ERROR) {
                  Tcl_SetResult(interp, "floating parameter error", NULL);
                  return TCL_ERROR;
               }
               *valptr++ = (int) &tmp->internalRep.doubleValue;
               Tcl_Free(tmp->bytes);
               tmp->bytes = NULL;
            }
            else {
               if (Tcl_GetIntFromObj(interp, tmp, valptr) == TCL_ERROR) {
                  Tcl_SetResult(interp, "integer parameter error", NULL);
                  return TCL_ERROR;
               }
               *valptr++ = (int) &tmp->internalRep.longValue;
               Tcl_Free(tmp->bytes);
               tmp->bytes = NULL;
            }
         }
         else if (type & STRING) {
            Tcl_GetString(tmp);
            *valptr++ = (int) &tmp->bytes;
         }
         else {
            if ((vptr = Tcl_GetByteArrayFromObj(tmp, 0)) == NULL) {
               Tcl_SetResult(interp, "binary string is required for \"void *\"", NULL);
               return TCL_ERROR;
            }
            *valptr++ = (int) vptr;
         }
      }
   }

   ret_type = params[3];
   dll_call((char *) valstrt, (char *)valptr - (char *)valstrt,
            (ret_type & PTR || ret_type & STRING) ? INT : ret_type & 0x0000003F,
            values, (void *) params[1], memcpy);

   error_code = GetLastError();
   Tcl_SetVar2Ex(interp, DLL "::error", NULL, Tcl_NewIntObj(error_code), 0);

   if (ret_type < PTR) {
      if (ret_type & NUMERIC) {
         if (ret_type & (FLOAT || DOUBLE)) {
            Tcl_SetObjResult(interp, Tcl_NewDoubleObj(*((double *) values)));
         }
         else {
            Tcl_SetObjResult(interp, Tcl_NewIntObj(*values));
         }
      }
      else if (ret_type & STRING) {
         Tcl_SetObjResult(interp, Tcl_NewStringObj((char *) values, -1));
      }
      else {
         Tcl_ResetResult(interp);
      }
   }
   else {
      Tcl_SetObjResult(interp, Tcl_NewIntObj(*values));

   }

   return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 */

int
dllBufferObjCmd(clientData, interp, objc, objv)
   ClientData clientData;			/*   */
   Tcl_Interp *interp;			/* Current interpreter. */
   int objc;				/* Number of arguments. */
   Tcl_Obj *CONST objv[];		/* Argument objects. */
{
   int size;
   char *memory;
   Tcl_Obj* tmp;


   if (objc != 2) {
      Tcl_WrongNumArgs(interp, 1, objv, "name size");
      return TCL_ERROR;
   }

   if (Tcl_GetIntFromObj(interp, objv[1], &size) == TCL_ERROR) {
      Tcl_SetResult(interp, "size should be an integer value", NULL);
      return TCL_ERROR;
   }
   memory = Tcl_Alloc(size);
   memset(memory, 0, size);
   tmp = Tcl_NewByteArrayObj(memory, size);
   Tcl_SetObjResult(interp, tmp);
   Tcl_Free(memory);
   return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 */

int
dllRefObjCmd(clientData, interp, objc, objv)
   ClientData clientData;			/*   */
   Tcl_Interp *interp;			/* Current interpreter. */
   int objc;				/* Number of arguments. */
   Tcl_Obj *CONST objv[];		/* Argument objects. */
{
   Tcl_Obj* tmp;


   if (objc !=2) {
      Tcl_WrongNumArgs(interp, 1, objv, "varname");
      return TCL_ERROR;
   }

   if ((tmp = Tcl_ObjGetVar2(interp, objv[1], NULL, 0)) == NULL) {
      tmp = Tcl_ObjSetVar2(interp, objv[1], NULL, Tcl_NewByteArrayObj(NULL, 0), 0);
   }

   if (strcmp(tmp->typePtr->name, "bytearray") != 0) {
   Tcl_SetResult(interp, "bytearray variable required", NULL);
   return TCL_ERROR;
   }

   Tcl_IncrRefCount(tmp);
   Tcl_Free(tmp->bytes);
   tmp->bytes = NULL;
   tmp = Tcl_NewIntObj((int) tmp->internalRep.twoPtrValue.ptr1);
   Tcl_SetObjResult(interp, tmp);
   return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 */

int
dllDeRefObjCmd(clientData, interp, objc, objv)
   ClientData clientData;			/*   */
   Tcl_Interp *interp;			/* Current interpreter. */
   int objc;				/* Number of arguments. */
   Tcl_Obj *CONST objv[];		/* Argument objects. */
{
   Tcl_Obj* tmp;


   if (objc !=2) {
      Tcl_WrongNumArgs(interp, 1, objv, "varname");
      return TCL_ERROR;
   }
   if ((tmp = Tcl_ObjGetVar2(interp, objv[1], NULL, 0)) == NULL) {
      Tcl_SetResult(interp, "parameter is not a variable", NULL);
      return TCL_ERROR;
   }
   if (strcmp(tmp->typePtr->name, "bytearray") != 0) {
      Tcl_SetResult(interp, "bytearray variable required", NULL);
      return TCL_ERROR;
   }
   Tcl_DecrRefCount(tmp);
   Tcl_ResetResult(interp);
   return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 */

int
dllMemRdObjCmd(clientData, interp, objc, objv)
   ClientData clientData;			/*   */
   Tcl_Interp *interp;			/* Current interpreter. */
   int objc;				/* Number of arguments. */
   Tcl_Obj *CONST objv[];		/* Argument objects. */
{
   int addr, lgt;


   if (objc !=2 && objc !=3) {
      Tcl_WrongNumArgs(interp, 1, objv, "memaddr ?length?");
      return TCL_ERROR;
   }

   if (Tcl_GetIntFromObj(interp, objv[1], &addr) == TCL_ERROR) {
   Tcl_SetResult(interp, "integer value for memory address is required", NULL);
   return TCL_ERROR;
   }
   if (objc == 3) {
      if (Tcl_GetIntFromObj(interp, objv[2], &lgt) == TCL_ERROR) {
      Tcl_SetResult(interp, "integer value for memory address is required", NULL);
      return TCL_ERROR;
      }
      if (lgt < 0) {
         Tcl_SetObjResult(interp, Tcl_NewStringObj((char *) addr, -1));
      }
      else {
         Tcl_SetObjResult(interp, Tcl_NewByteArrayObj((char *) addr, lgt));
      }
   }
   else {
         memcpy((char *) &lgt, (char *) addr, 4);
         addr += 4;
         Tcl_SetObjResult(interp, Tcl_NewByteArrayObj((char *) addr, lgt));
   }
   return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 */

int
dllMemWrObjCmd(clientData, interp, objc, objv)
   ClientData clientData;			/*   */
   Tcl_Interp *interp;			/* Current interpreter. */
   int objc;				/* Number of arguments. */
   Tcl_Obj *CONST objv[];		/* Argument objects. */
{
   int addr, lgt;
   char *str;

   if (objc !=3) {
      Tcl_WrongNumArgs(interp, 1, objv, "bytearray memaddr");
      return TCL_ERROR;
   }
   if (strcmp(objv[1]->typePtr->name, "bytearray") != 0) {
     Tcl_SetResult(interp, "bytearray variable required", NULL);
     return TCL_ERROR;
   }
   if (Tcl_GetIntFromObj(interp, objv[2], &addr) == TCL_ERROR) {
      Tcl_SetResult(interp, "integer value for memory address is required", NULL);
   return TCL_ERROR;
   }
   str = Tcl_GetByteArrayFromObj(objv[1], &lgt);
   memcpy((char *) addr, str, lgt);
   return TCL_OK;
}


