#Copyright Jo'zsef Ne'meth 2003, joe.nemeth@cpluscsystems.axelero.net

#Edited by robin 2011 (homay2-at-freemail.hu)
#- Store dll.dll alongside this script
#- Optimized for being integrated in tclkitsh.exe (or other Starpack/Starkit).
#  The loading of the .dll checks if we're in a VFS (Starpack/Starkit case) and
#  if yes, it composes a permanent path for the .dll under %TEMP% and extracts
#  the .dll there (if it's not there yet) and loads from there. This avoids the
#  extracting-deleting steps from occurring every time when the .dll is first
#  used in a process, and Windows' prefetcher / Superfetch can speed up the
#  loading of this .dll, too.
#  However, you should be aware of this technique because it is also a security
#  hole: malware may infect the formerly extracted .dll in %TEMP%!

# After loading dll.dll, Dll_Init creates a namespace and
# stores its name in the global variable DLL.

namespace eval ::tcl_dllcaller_1103090332 {
  set dll [file join [file dirname [info script]] dll]
  set err 1
  foreach f [list $dll.dll $dll.so] {
    if {![file exists $f]} continue
    if {[lindex [file system $f] 0] ne "native"} {
       set d [string map {: ""} [namespace current]]
       set d [file join $env(TMP) $d [file tail $f]]
       if {![file exists $d] || [file size $d]!=[file size $f]} {
          file mkdir [file dirname $d]
          file copy -force -- $f $d 
       }
       set f $d
    }
    set err [catch {load $f}]
    if {!$err} break
  }
  if {$err} { error "Cannot find $dll.\[dll|so\]" }
  namespace delete [namespace current]
}

package provide dllcaller 1.0
namespace eval ${::DLL} {

  namespace export load buffer memrd memwr ref deref

#define void	0x00000000
#define char	0x00000001
#define short	0x00000002
#define int		0x00000004
#define int64	0x00000008
#define float	0x00000010
#define double	0x00000020
#define unsigned	0x00000040
#define numeric	0x00000080
#define string	0x00000100
#define voidret	0x10000000
#define ptr		0x40000000
#define ptrptr	0x80000000
array set type {char 129 short 130 int 132 int64 136 float 144 double 160}
array set type {uchar 193 ushort 194 uint 196 uint64 200 string 256}
array set type {void 0 "" 0 bufret 268435456 * 1073741824 ** 2147483648}

# Create a flag word from a base type and pointer specification
proc _getFlag_ {base ptr} {
   variable type

   set flag $type($base)
   if {$flag != 129} {
      set flag [expr {$flag + $type($ptr)}]
   } elseif {$ptr ne ""} {
      set flag [expr {$flag + $type($ptr) - 1073741697}]
   }
   return [binary format i $flag]
}

# Create a command from a dynamic library function
# Usage: cmd "int GetDesktopWindow()" [-> alias]
# Creates 'alias' proc, by default in $nsp. 'alias' defaults to function name
proc cmd {nsp def} {
   variable type
   variable entry
   variable alias

   set length [llength $def]
   if {$length != 1 && $length != 3} {
      error "Wrong number of arguments"
   }
   set handle [set ${nsp}::handle]
   set spec [lindex $def 0]
   set defList [regexp -inline -- {^(.*) +([_A-Za-z0-9]+) *\((.*) *$} $spec]
   set name [lindex $defList 2]
   if {$name eq ""} {error "Syntax error"}

   set retspec  [lindex $defList 1]
   if {$retspec eq ""} {error "Return type must be specified"}
   set retval [regexp -inline -- \
                  {^ *(u?char|u?short|u?int|u?int64|void(/[0-9]+)?|float|double) *(\*{0,2}) *} \
                  $retspec]
   if {[lindex $retval 2] ne ""} {
      set params [expr $type(bufret) + $type(*) + [string range [lindex $retval 2] 1 end]]
   } else {
      set params [_getFlag_ [lindex $retval 1] [lindex $retval 3]]
   }

   set spec [lindex $defList 3]
   set parlist [regexp -inline -all -- \
                  { *(u?char|u?short|u?int|u?int64|void|float|double) *(\*{0,2}) *[,)]} \
                  $spec]

   set cnt 1
   foreach {m t p} $parlist {
      incr cnt
      set params $params[_getFlag_ $t $p]
   }

   if {$length == 3} {
      if {[lindex $def 1] ne "->"} {
         error "Syntax error"
      }
      set _alias [lindex $def 2]
      if {[namespace tail $_alias] eq $_alias} {
         set _alias [namespace_join $nsp $_alias]
      } elseif {[namespace qualifiers $_alias] eq ""} {
         set _alias [namespace tail $_alias]
      } else {
         error "Invalid alias specification"
      }
   } else {
      set _alias [namespace_join $nsp $name]
   }
   set address [create_cmd $_alias $handle $name]
   if {!$address} {
      error "Dynamic library function $name not found"
   }
   set entry($_alias) \
       [binary format i $handle][binary format i $address][binary format i $cnt]$params[buffer [expr 8 * $cnt]]
   set alias($_alias) [namespace_join $nsp $name]
}

#Load a dynamic library and register its handle
#Creates the namespace 'alias'
#Usage: load user32 [-> alias]
proc load {args} {

   set length [llength $args]
   if {$length != 1 && $length != 3} {
      error "Wrong number of arguments"
   }

   set name [file rootname [file tail [lindex $args 0]]]

   if {$length == 3} {
      if {[lindex $args 1] ne "->"} {
         error "Syntax error"
      }
      set alias [lindex $args 2]
   } else {
      set alias $name
   }
   set handle [load_module [lindex $args 0]]
   if {!$handle} {
      error "Dynamic library not found"
   }

   set alias [namespace_join :: $alias]
   namespace eval $alias proc cmd args \
      "{namespace inscope [namespace current] cmd $alias \$args}"

   set ${alias}::handle $handle
   set ${alias}::alias $name
}

proc namespace_join {a b} {
  if {"::" == [string range $a end-1 end]} {
     set a [string range $a 0 end-2]
  }
  if {"::" == [string range $b 0 1]} {
     set b [string range $b 2 end]
  }
  return "${a}::$b"
}

if {[namespace current] ne "::dllcaller"} {
   namespace eval ::dllcaller namespace import [namespace current]::*
}

}
